/**************************************************************************
 *
 * text_t.cpp -- a simple 16-bit charater string class
 * Copyright (c) 1998 -- Rodger McNab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: text_t.cpp 4 1998-11-17 01:10:39Z sjboddie $
 *
 **************************************************************************/

/*
   $Log$
   Revision 1.1  1998/11/17 01:11:06  sjboddie
   Initial revision

   Revision 1.7  1998/06/24 01:50:08  rjmcnab
   Added stuff for the objectspace libraries

   Revision 1.6  1998/06/22 22:11:14  rjmcnab
   Added copy operator = for ints and c-strings for efficiency
   Added some general functions for dealing with text_ts

   Revision 1.5  1998/06/18 23:53:52  rjmcnab
   added '+' operator for concatenation
   removed global declaration of ascii2text_t and text_t2ascii
     (too unfriendly for threads)
   added support for integers

   Revision 1.4  1998/06/17 04:52:32  rjmcnab
   added support for arrays of characters

   Revision 1.3  1998/06/17 02:09:42  rjmcnab
   added getcstr

   Revision 1.2  1998/06/15 22:01:19  rjmcnab
   fixed a few bugs

   Revision 1.1  1998/06/13 20:08:34  rjmcnab
   Initial revision

*/

static char *RCSID = "$Id: text_t.cpp 4 1998-11-17 01:10:39Z sjboddie $";


#ifndef USE_OBJECTSPACE
#  include <algorithm>
#else
#  include <ospace\std\algorithm>
#endif


#include "text_t.h"

////////////////////////////////////
// text_t methods
////////////////////////////////////

text_t::text_t () 
{
  clear ();
}

text_t::text_t (int i) 
{
  clear ();
  appendint (i);
}

text_t::text_t (char *s) 
{
  clear ();
  appendcstr (s);
}

void text_t::append (const text_t &t) 
{
  const_iterator here, end=t.end();
  for (here=t.begin(); here!=end;here++)
    {
      text.push_back(*here);
    }
}

void text_t::appendrange (iterator first, iterator last) 
{
  while (first != last) 
    {
      text.push_back (*first);
      first++;
    }
}

void text_t::appendrange (const_iterator first, const_iterator last) 
{
  while (first != last) 
    {
      text.push_back (*first);
      first++;
    }
}

void text_t::appendint (int i)
{
  // deal with zeros and negatives
  if (i == 0) 
    {
      text.push_back('0');
      return;
    }
  else if (i < 0)
    {
      text.push_back('-');
      i *= -1;
    }

  // get a buffer for the conversion
  int maxbuflen = sizeof(int)*3;
  char *buf = new char[maxbuflen];
  int len = 0;
  
  // get the number in reverse
  while (i > 0)
    {
      buf[len++] = '0'+ (i%10);
      i = i/10;
    }

  // reverse the number
  while (len > 0)
    {
      text.push_back(buf[--len]);
    }

  delete buf;
}

int text_t::getint ()
{
  int i = 0;
  int mult = 1; // become -1 for negative numbers

  iterator here = text.begin();
  iterator end = text.end();
  
  // do plus and minus signs
  if (here != end)
    {
      if (*here == '-')
	{
	  mult = -1;
	  here++;
	}
      else if (*here == '+')
	{
	  mult = 1;
	  here++;
	}
    }

  // deal with the number
  while ((here != end) && (*here >= '0') && (*here <= '9'))
    {
      i = 10*i + (*here - '0');
      here++;
    }

  i *= mult;
  return i;
}



void text_t::appendcarr (char *s, size_type len)
{
  unsigned char *us = (unsigned char *)s;
  while (len > 0) 
    {
      text.push_back (*us); // append this character
      us++;
      len--;
    }
}

void text_t::appendcstr (char *s) 
{
  unsigned char *us = (unsigned char *)s;
  while (*us != '\0') 
    {
      text.push_back (*us); // append this character
      us++;
    }
}


// strings returned from getcarr and getcstr become the callers 
// responsibility and should be deallocated with "delete"

char *text_t::getcarr(size_type &len) const
{
  unsigned char *cstr = new unsigned char[size()];
  const_iterator ithere = begin();
  const_iterator itend = end();

  while (ithere != itend)
    {
      if (*ithere >= 256) cstr[len] = ' ';
      else cstr[len] = (*ithere);
      len++;
      ithere++;
    }

  return (char *)cstr;
}

char *text_t::getcstr() const
{
  unsigned char *cstr = new unsigned char[size() + 1];
  const_iterator ithere = begin();
  const_iterator itend = end();
  int len = 0;

  while (ithere != itend)
    {
      if (*ithere >= 256) cstr[len] = ' ';
      else cstr[len] = (*ithere);
      len++;
      ithere++;
    }

  cstr[len] = '\0';

  return (char *)cstr;
}


// general functions which work on text_ts

// find a character within a range
text_t::const_iterator findchar (text_t::const_iterator first, text_t::const_iterator last, 
				 unsigned short c)
{
  while (first != last)
    {
      if (*first == c) break;
      first++;
    }
  return first;
}

text_t::iterator findchar (text_t::iterator first, text_t::iterator last, 
			   unsigned short c)
{
  while (first != last)
    {
      if (*first == c) break;
      first++;
    }
  return first;
}

// get a string up to the next delimiter (which is skipped)
text_t::const_iterator getdelimitstr (text_t::const_iterator first, 
				      text_t::const_iterator last,
				      unsigned short c, text_t &outstr)
{
  text_t::const_iterator here = first;
  here = findchar (first, last, c);
  outstr.clear();
  outstr.appendrange (first, here);
  if (here != last) here++; // skip c
  return here;
}

text_t::iterator getdelimitstr (text_t::iterator first, text_t::iterator last,
				unsigned short c, text_t &outstr)
{
  text_t::iterator here = first;
  here = findchar (first, last, c);
  outstr.clear();
  outstr.appendrange (first, here);
  if (here != last) here++; // skip c
  return here;
}

// split a string with a character
void splitchar (text_t::const_iterator first, text_t::const_iterator last,
		unsigned short c, text_tlist &outlist)
{
  outlist.erase(outlist.begin(), outlist.end());

  text_t t;

  while (first != last)
    {
      first = getdelimitstr (first, last, c, t);
      outlist.push_back (t);
    }
}

void splitchar (text_t::const_iterator first, text_t::const_iterator last,
		unsigned short c, text_tarray &outlist)
{
  outlist.erase(outlist.begin(), outlist.end());

  text_t t;

  while (first != last)
    {
      first = getdelimitstr (first, last, c, t);
      outlist.push_back (t);
    }
}

// join a string using a character
void joinchar (const text_tlist &inlist, unsigned short c, text_t &outtext)
{
  outtext.clear ();

  text_tlist::const_iterator here = inlist.begin ();
  text_tlist::const_iterator end = inlist.end ();
  bool first = true;
  while (here != end)
    {
      if (!first) outtext.push_back (c);
      first = false;
      outtext += *here;
      here++;
    }
}

void joinchar (const text_tarray &inlist, unsigned short c, text_t &outtext)
{
  outtext.clear ();

  text_tarray::const_iterator here = inlist.begin ();
  text_tarray::const_iterator end = inlist.end ();
  bool first = true;
  while (here != end)
    {
      if (!first) outtext.push_back (c);
      first = false;
      outtext += *here;
      here++;
    }
}

// count the occurances of a character within a range
int countchar (text_t::const_iterator first, text_t::const_iterator last,
	       unsigned short c)
{
  int count = 0;
  while (first != last) {
    if (*first == c) count ++;
    first ++;
  }
  return count;
}



////////////////////////////////////
// convertclass methods
////////////////////////////////////

// conversion classes used for getting information in to and out of
// the text_t class.

convertclass::convertclass () 
{
  // nothing to do
}

void convertclass::reset ()
{
  // nothing to do
}


////////////////////////////////////
// inconvertclass methods
////////////////////////////////////

// convert from a char stream to the text_t class
// the default version assumes the input is a ascii
// character array

inconvertclass::inconvertclass () 
{
  start = NULL;
  len = 0;
}


void inconvertclass::reset ()
{
  start = NULL;
  len = 0;
}

void inconvertclass::setinput (char *thestart, size_t thelen)
{
  start = thestart;
  len = thelen;
}

void inconvertclass::convert (text_t &output, status_t &status)
{
  if (start == NULL || len == 0)
    {
      status = finished;
      return;
    }

  // don't want any funny sign conversions happening
  unsigned char *here = (unsigned char *)start;
  while (len > 0) 
    {
      output.push_back (*here); // append this character
      ++here;
      --len;
    }

  start = (char *)here; // save current position
  status = finished;
}

// an instance of the default inconvertclass to do simple
// conversions. Note that any functions that use this are
// not reentrant. If a function needs to be reentrant it
// should declare its own instance.
inconvertclass ascii2text_t;


////////////////////////////////////
// outconvertclass methods
////////////////////////////////////

// Convert from a text_t class to a char stream
// This default version assumes the output is a ascii
// character array. If you set the output stream you
// can use this class to output to a stream using the
// << operator. The << operator can also be conveniently
// used to set the output stream by doing something like
//
// cout << text_t2ascii << text_tstr << anothertext_tstr;
//
outconvertclass::outconvertclass ()
{
  input = NULL;
  outs = NULL;
}

void outconvertclass::reset ()
{
  input = NULL;
  outs = NULL;
}

void outconvertclass::setinput (text_t *theinput)
{
  input = theinput;
  if (input != NULL) texthere = input->begin();
}

void outconvertclass::convert (char *output, size_t maxlen, 
		      size_t &len, status_t &status)
{
  if (input == NULL || output == NULL)
    {
      status = finished;
      return;
    }

  // don't want any funny sign conversions happening
  unsigned char *uoutput = (unsigned char *)output;
  text_t::iterator textend = input->end();
  len = 0;
  while ((len < maxlen) && (texthere != textend)) 
    {
      if (*texthere < 256) *uoutput = (unsigned char)(*texthere);
      else *uoutput = 32; // put a space where a char >= 256 exists
      ++uoutput;
      ++len;
      ++texthere;
    }
  
  if (texthere == textend) status = finished;
  else status = unfinished;
}

void outconvertclass::setostream (ostream *theouts)
{
  outs = theouts;
}

ostream *outconvertclass::getostream ()
{
  return outs;
}




// an instance of the default outconvertclass to do simple
// conversions
outconvertclass text_t2ascii;



// stream operators for the output class

outconvertclass &operator<< (ostream &theouts, outconvertclass &outconverter)
{
  outconverter.setostream(&theouts);
  return outconverter;
}


#define STREAMBUFSIZE 256
outconvertclass &operator<< (outconvertclass &outconverter, const text_t &t)
{
  ostream *outstream = outconverter.getostream();

  if (outstream == NULL) return outconverter;

  char outbuf[STREAMBUFSIZE];
  size_t len;
  outconvertclass::status_t status = outconvertclass::unfinished;

  // assume that there is no data needing converting
  // left in the converter
  outconverter.setinput ((text_t *)(&t)); // note the const -> nonconst conversion

  while (status == outconvertclass::unfinished)
    {
      outconverter.convert (outbuf, STREAMBUFSIZE, len, status);
      if (len > 0) outstream->write(outbuf, len);
    }

  return outconverter;
}
